/*
 * Copyright (c) 2009, 2010
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id$ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include <cforce.h>

#include "acld.h"
#include "cforcep.h"

/* Locals */

/* Forwards */
const char *
cforceerrstr(int);

const char *
cforceerrstr(int e)
{
	static char buf[32];

	switch (e) {

	case ERROR_GENERAL:
		return ("GENERAL");

	case ERROR_COMMUNICATION:
		return ("COMMUNICATION");

	case ERROR_INVALID_IP_ADDRESS:
		return ("INVALID_IP_ADDRESS");

	case ERROR_ADDRESS_ALREADY_PROVISIONED:
		return ("ADDRESS_ALREADY_PROVISIONED");

	case ERROR_HOST_ALREADY_BLOCKED:
		return ("HOST_ALREADY_BLOCKED");

	case ERROR_INVALID_ID:
		return ("INVALID_ID");

	case ERROR_CAPACITY_EXCEEDED:
		return ("CAPACITY_EXCEEDED");

	case ERROR_INVALID_RULE:
		return ("INVALID_RULE");

	case ERROR_BOTH_NOT_SUPPORTED:
		return ("BOTH_NOT_SUPPORTED");

	case ERROR_CONVERSATION_ALREADY_BLOCKED:
		return ("CONVERSATION_ALREADY_BLOCKED");

	case ERROR_NOT_INITIALIZED:
		return ("NOT_INITIALIZED");

	case ERROR_INVALID_ARGUMENTS:
		return ("INVALID_ARGUMENTS");

	case ERROR_INVALID_CONNECTION_PARAMETERS:
		return ("INVALID_CONNECTION_PARAMETERS");

	case ERROR_ALL_RULE_PROVISIONED:
		return ("ALL_RULE_PROVISIONED");

	case ERROR_RULE_CONFLICT:
		return ("RULE_CONFLICT");

	case ERROR_INVALID_COMBINATION:
		return ("INVALID_COMBINATION");

	case ERROR_COULD_NOT_OPEN_FILE:
		return ("COULD_NOT_OPEN_FILE");

	case ERROR_COULD_NOT_WRITE_TO_FD:
		return ("COULD_NOT_WRITE_TO_FD");

	case ERROR_DATA_STRUCTURE_INCONSISTENCY:
		return ("DATA_STRUCTURE_INCONSISTENCY");

	default:
		break;
	}
	(void)snprintf(buf, sizeof(buf), "#%d", e);
	return (buf);
}

int
cforceinit(struct state *sp)
{
	int rc;

	rc = cforce_initialize(sp->cf->c_cforceaddr, sp->cf->c_portcforce);
	if (rc < 0) {
		lg(LOG_ERR, "cforceinit: cforce_initialize failed: %s",
		    cforceerrstr(rc));
		return (-1);
	}

	// Zero rules

	// reinstall saved rules

	// done

	++sp->cforce;
	return (0);
}
